"""Load NYU Depth V2 data.

Downloaded NYUDepth dataset link:
    https://drive.google.com/drive/folders/1TzwfNA5JRFTPO-kHMU___kILmOEodoBo
This dataset is not the official NYU Depth V2 datset.
The data are preprocessed.
For more information, see the link below:
    https://github.com/ialhashim/DenseDepth
"""
import logging
import os
from zipfile import ZipFile

import pandas as pd
from PIL import Image

import datasetinsights.constants as const
from datasetinsights.io.gcs import GCSClient

from .base import Dataset

NYU_GCS_PATH = "data/nyudepth"
NYUDEPTH_LOCAL_PATH = "nyudepth"
SPLITS = ["train", "val", "test"]
ZIPFILE = "nyu_v2_subset_data.zip"
UNZIP_NAME = "data"
REAL_IMAGE_HEADER = "real"
TARGET_IMAGE_HEADER = "target"

logger = logging.getLogger(__name__)


class NyuDepth(Dataset):
    """
    Attributes:
        root (str): local directory where nyu_v2 dataset is saved
        nyu2_data (Pandas DataFrame): training or testing data
        split (str): test or valuation. see SPLITS constant for possible values
    """

    def __init__(
        self,
        data_root=const.DEFAULT_DATA_ROOT,
        transforms=None,
        split="train",
        cloud_path=NYU_GCS_PATH,
    ):
        """
        Args:
            data_root (str): root directory prefix of all datasets
            split (str): indicate split type of the dataset (train|test)
            cloud_path (str): cloud storage path where a copy of nyu_v2 is
            saved
        """
        self.root = os.path.join(data_root, NYUDEPTH_LOCAL_PATH)
        self.transforms = transforms
        self.split = split
        if not os.path.isdir(self.root):
            os.mkdir(self.root)
        if split not in SPLITS:
            raise ValueError(
                f"invalid value for split: {split}, possible values "
                f"are: {SPLITS}"
            )
        self.download(cloud_path=cloud_path)

        unzip_dir = os.path.join(self.root, UNZIP_NAME)

        self.nyu2_data = pd.read_csv(
            os.path.join(unzip_dir, f"nyu2_{split}.csv"),
            header=None,
            names=[REAL_IMAGE_HEADER, TARGET_IMAGE_HEADER],
            skip_blank_lines=True,
        )

    def __getitem__(self, index):
        """
        Args:
            index (int): index of the element
        Returns:
            sample (tuple): (real image, depth target)
        """
        real_img_path = self.nyu2_data[REAL_IMAGE_HEADER][index]
        target_img_path = self.nyu2_data[TARGET_IMAGE_HEADER][index]
        real_img = Image.open(os.path.join(self.root, real_img_path))
        target_img = Image.open(os.path.join(self.root, target_img_path))
        sample = (real_img, target_img)
        if self.transforms:
            sample = self.transforms(sample)
        return sample

    def __len__(self):
        return len(self.nyu2_data)

    def download(self, cloud_path):
        """Download nyu_v2 dataset
        The directory structure of the downloaded data is
        |--self.root
           |--nyudepth
               |--nyu_data.zip
               |--data
                   |--nyu2_test.csv
                   |--nyu2_test
                         |--00000_colors.png
                         |--00000_depth.png ...
                         |--01448_colors.png
                         |--01448_depth.png
                   |--nyu2_train.csv
                   |--nyu2_train
                         |--basement_0001a_out
                              |--1.jpg
                              |--1.png ...
                              |--281.jpg
                              |--281.png
                         ...
                         |--study_room_0005b_out
                              |--1.jpg
                              |--1.png ...
                              |--133.jpg
                              |--133.png
        Args:
            cloud_path (str): cloud path of the dataset
        """
        zip_file = os.path.join(self.root, ZIPFILE)
        unzip_dir = os.path.join(self.root, UNZIP_NAME)

        if os.path.isfile(zip_file):
            logger.debug(f"File {zip_file} exists. Skip download.")
        else:
            client = GCSClient()
            object_key = os.path.join(NYU_GCS_PATH, ZIPFILE)

            logger.debug(
                f"Downloading file {zip_file} from gs://{const.GCS_BUCKET}/"
                f"{object_key}"
            )
            client.download(
                local_path=self.root, bucket=const.GCS_BUCKET, key=object_key,
            )

        if os.path.isdir(unzip_dir):
            logger.debug(f"File {unzip_dir} exists. Skip unzip.")
        else:
            # unzip the file
            with ZipFile(zip_file, "r") as zip_ref:
                zip_ref.extractall(self.root)
                logger.debug(f"Unzip file from {zip_file}")



from .base import Estimator, create_estimator
from .deeplab import DeeplabV3
from .densedepth import DenseDepth
from .faster_rcnn import FasterRCNN, convert_bboxes2canonical

__all__ = [
    "Estimator",
    "create_estimator",
    "DeeplabV3",
    "DenseDepth",
    "FasterRCNN",
    "convert_bboxes2canonical",
]



import logging
import os
from pathlib import Path

from torchvision import datasets

import datasetinsights.constants as const
from datasetinsights.io.gcs import GCSClient

from .base import Dataset

CITYSCAPES_GCS_PATH = "data/cityscapes"
CITYSCAPES_LOCAL_PATH = "cityscapes"
ZIPFILES = [
    "leftImg8bit_trainvaltest.zip",
    "gtFine_trainvaltest.zip",
]
CITYSCAPES_COLOR_MAPPING = {c.id: c.color for c in datasets.Cityscapes.classes}

logger = logging.getLogger(__name__)


class Cityscapes(Dataset):
    """
    Args:
        data_root (str): root directory prefix of all datasets
        split (str): indicate split type of the dataset (train|val|test)

    Attributes:
        root (str): root directory of the dataset
        split (str): indicate split type of the dataset (train|val|test)
    """

    def __init__(
        self,
        *,
        data_root=const.DEFAULT_DATA_ROOT,
        split="train",
        transforms=None,
        **kwargs,
    ):
        self.root = os.path.join(data_root, CITYSCAPES_LOCAL_PATH)
        self.split = split
        self.download(CITYSCAPES_GCS_PATH)

        self._cityscapes = datasets.Cityscapes(
            self.root,
            split=split,
            mode="fine",
            target_type="semantic",
            transforms=transforms,
        )

    def __getitem__(self, index):
        return self._cityscapes[index]

    def __len__(self):
        return len(self._cityscapes)

    def download(self, cloud_path):
        """Download cityscapes dataset

        Note:
            The current implementation assumes a GCS cloud path.
            Should we keep this method here if we want to support other cloud
            storage system?

        Args:
            cloud_path (str): cloud path of the dataset
        """
        path = Path(self.root)
        path.mkdir(parents=True, exist_ok=True)

        for zipfile in ZIPFILES:
            localfile = os.path.join(self.root, zipfile)
            if os.path.isfile(localfile):
                # TODO: Check file hash to verify file integrity
                logger.debug(f"File {localfile} exists. Skip download.")
                continue
            client = GCSClient()
            object_key = os.path.join(CITYSCAPES_GCS_PATH, zipfile)

            logger.debug(
                f"Downloading file {localfile} from gs://{const.GCS_BUCKET}/"
                f"{object_key}"
            )
            client.download(
                local_path=self.root, bucket=const.GCS_BUCKET, key=object_key
            )



r"""Average Log10 Error metric.

The average log10 error can be described as:

.. math::
    \frac{1}{n}\sum_{p}^{n} |log_{10}(y_p)-log_{10}(\hat{y_p})|

"""
import numpy as np

from .base import EvaluationMetric
from .exceptions import NoSampleError


class AverageLog10Error(EvaluationMetric):
    """Average Log10 Error metric.

    The metric is defined for grayscale depth images.

    Attributes:
        sum_of_log10_error (float): the sum of the log10 errors for all the
        images in a branch
        num_samples (int): the number of samples in all mini-batches
    """

    def __init__(self):
        self.sum_of_log10_error = 0
        self.num_samples = 0

    def reset(self):
        self.sum_of_log10_error = 0
        self.num_samples = 0

    def update(self, output):
        y_pred, y = output
        if y.shape != y_pred.shape:
            raise ValueError(
                f"The shapes of real data {y.shape} and predicted data "
                f"{y_pred.shape} should be the same."
            )

        self.num_samples += y.shape[0]
        self.sum_of_log10_error += (
            abs(np.log10(y + 1e-15) - np.log10(y_pred + 1e-15)).mean().item()
            * y.shape[0]
        )

    def compute(self):
        if self.num_samples == 0:
            raise NoSampleError(
                "AverageLog10Error must have at least one example before it "
                "can be computed."
            )
        return self.sum_of_log10_error / self.num_samples



r"""Average Relative Error metrics.

The average relative error can be described as:

.. math::
    \frac{1}{num\ samples}\sum_{p}^{num\ samples}\frac{|y_p-\hat{y_p}|}{y_p}
"""
import numpy as np

from .base import EvaluationMetric
from .exceptions import NoSampleError


class AverageRelativeError(EvaluationMetric):
    """Average Relative Error metric.

    The metric is defined for grayscale depth images.

    Attributes:
        sum_of_relative_error (float): the sum of the relative errors for all
        the images in a branch
        num_samples (int): the number of samples in all mini-batches
    """

    def __init__(self):
        self.sum_of_relative_error = 0
        self.num_samples = 0

    def reset(self):
        self.sum_of_relative_error = 0
        self.num_samples = 0

    def update(self, output):
        y_pred, y = output
        if y.shape != y_pred.shape:
            raise ValueError(
                f"The shapes of real data {y.shape} and predicted data "
                f"{y_pred.shape} should be the same."
            )

        self.num_samples += y.shape[0]
        self.sum_of_relative_error += (
            np.abs(y - y_pred) / (y + 1e-15)
        ).mean().item() * y.shape[0]

    def compute(self):
        if self.num_samples == 0:
            raise NoSampleError(
                "AverageRelativeError must have at least one example before"
                " it can be computed."
            )
        return self.sum_of_relative_error / self.num_samples



TP = True
FP = False


class Records:
    """Save prediction records during update.

    Attributes:
        iou_threshold (float): iou threshold
        match_results (list): save the results (TP/FP)
    Args:
        iou_threshold (float): iou threshold (default: 0.5)
    """

    def __init__(self, iou_threshold=0.5):
        self.iou_threshold = iou_threshold
        self.match_results = []

    def reset(self):
        self.match_results = []

    def add_records(self, gt_bboxes, pred_bboxes):
        """Add ground truth and prediction records.

        Args:
            gt_bboxes: ground truth bboxes in the current image
            pred_bboxes: sorted predicition bboxes in the current image
        """
        gt_seen = [False] * len(gt_bboxes)

        for pred_bbox in pred_bboxes:
            max_iou = -1
            max_idx = -1
            for i, gt_bbox in enumerate(gt_bboxes):
                if gt_bbox.label != pred_bbox.label:
                    continue
                iou = gt_bbox.iou(pred_bbox)
                if iou > max_iou:
                    max_iou = iou
                    max_idx = i
            if max_iou >= self.iou_threshold:
                if not gt_seen[max_idx]:
                    gt_seen[max_idx] = True
                    self.match_results.append((pred_bbox.score, TP))
                else:
                    self.match_results.append((pred_bbox.score, FP))
            else:
                self.match_results.append((pred_bbox.score, FP))



from abc import ABCMeta, abstractmethod


class Dataset(metaclass=ABCMeta):
    """Abstract base class for datasets.
    """

    @staticmethod
    def create(name, **kwargs):
        """Create a new instance of the dataset subclass

        Args:
            name (str): unique identifier for a dataset subclass
            config (dict): parameters specific to each dataset subclass
                used to create a dataset instance

        Returns:
            an instance of the specified dataset subclass
        """
        dataset_cls = Dataset.find(name)

        return dataset_cls(**kwargs)

    @staticmethod
    def find(name):
        """Find Dataset subclass based on the given name

        Args:
            name (str): unique identifier for a dataset subclass

        Returns:
            a label of the specified dataset subclass
        """
        dataset_classes = Dataset.__subclasses__()
        dataset_names = [d.__name__ for d in dataset_classes]
        if name in dataset_names:
            dataset_cls = dataset_classes[dataset_names.index(name)]
            return dataset_cls
        else:
            raise NotImplementedError(f"Unknown Dataset class {name}!")

    @abstractmethod
    def __getitem__(self, index):
        raise NotImplementedError("Subclass needs to implement this method")

    @abstractmethod
    def __len__(self):
        raise NotImplementedError("Subclass needs to implement this method")



import logging
import os
import zipfile
from pathlib import Path
from typing import List, Tuple

import torch
import torchvision
from PIL.Image import Image

import datasetinsights.constants as const
from datasetinsights.io.bbox import BBox2D
from datasetinsights.io.gcs import GCSClient

from .base import Dataset

ANNOTATION_FILE_TEMPLATE = "{}_{}2017.json"
COCO_GCS_PATH = "data/coco"
COCO_LOCAL_PATH = "coco"
logger = logging.getLogger(__name__)


def _coco_remove_images_without_annotations(dataset):
    """

    Args:
        dataset (torchvision.datasets.CocoDetection):

    Returns (torch.utils.data.Subset): filters dataset to exclude examples
    which either have no bounding boxes or have an invalid bounding box (a
    bounding box is invalid if it's height or width is <1).

    """

    def _has_any_empty_box(anno):
        return any(any(o <= 1 for o in obj["bbox"][2:]) for obj in anno)

    def _has_valid_annotation(anno):
        # if it's empty, there is no annotation
        if len(anno) == 0:
            return False
        if _has_any_empty_box(anno):
            return False
        return True

    ids = []
    for ds_idx, img_id in enumerate(dataset.ids):
        ann_ids = dataset.coco.getAnnIds(imgIds=img_id, iscrowd=None)
        anno = dataset.coco.loadAnns(ann_ids)
        if _has_valid_annotation(anno):
            ids.append(ds_idx)
    dataset = torch.utils.data.Subset(dataset, ids)
    return dataset


def convert_coco2canonical(coco_annotation):
    """
    convert from a tuple of image and coco style dictionary describing the
    bboxes to a tuple of image, List of BBox2D
    Args:
        coco_annotation (tuple): image and coco style dictionary

    Returns: a tuple of image, List of BBox2D

    """
    image, targets = coco_annotation
    all_bboxes = []
    for t in targets:
        label = t["category_id"]
        bbox = t["bbox"]
        b = BBox2D(x=bbox[0], y=bbox[1], w=bbox[2], h=bbox[3], label=label)
        all_bboxes.append(b)
    return image, all_bboxes


class CocoDetection(Dataset):
    """
    http://cocodataset.org/#detection-2019
    """

    def __init__(
        self,
        *,
        data_root=const.DEFAULT_DATA_ROOT,
        split="train",
        transforms=None,
        remove_examples_without_boxes=True,
        **kwargs,
    ):
        # todo add test split
        self.split = split
        self.root = os.path.join(data_root, COCO_LOCAL_PATH)
        self.download()
        self.coco = self._get_coco(root=self.root, image_set=split)
        if remove_examples_without_boxes:
            self.coco = _coco_remove_images_without_annotations(
                dataset=self.coco
            )
        self.transforms = transforms

    def __getitem__(self, idx) -> Tuple[Image, List[BBox2D]]:
        """
        Args:
            idx:

        Returns: Image with list of bounding boxes found inside the image

        """
        image, target = convert_coco2canonical(self.coco[idx])
        if self.transforms:
            image, target = self.transforms(image, target)

        return image, target

    def __len__(self):
        return len(self.coco)

    def _get_coco(self, root, image_set, mode="instances"):
        PATHS = {
            "train": (
                "train2017",
                os.path.join(
                    "annotations",
                    ANNOTATION_FILE_TEMPLATE.format(mode, "train"),
                ),
            ),
            "val": (
                "val2017",
                os.path.join(
                    "annotations", ANNOTATION_FILE_TEMPLATE.format(mode, "val")
                ),
            ),
        }
        img_folder, ann_file = PATHS[image_set]
        img_folder = os.path.join(root, img_folder)
        ann_file = os.path.join(root, ann_file)
        coco = torchvision.datasets.CocoDetection(img_folder, ann_file)
        return coco

    def _get_local_annotations_zip(self):
        return os.path.join(self.root, "annotations_trainval2017.zip")

    def _get_local_images_zip(self):
        return os.path.join(self.root, f"{self.split}2017.zip")

    def download(self, cloud_path=COCO_GCS_PATH):
        path = Path(self.root)
        path.mkdir(parents=True, exist_ok=True)
        client = GCSClient()
        annotations_zip_gcs = f"{cloud_path}/annotations_trainval2017.zip"
        annotations_zip_2017 = self._get_local_annotations_zip()
        logger.info(f"checking for local copy of data")
        if not os.path.exists(annotations_zip_2017):
            logger.info(f"no annotations zip file found, will download.")
            client.download(
                local_path=self.root,
                bucket=const.GCS_BUCKET,
                key=annotations_zip_gcs,
            )
            with zipfile.ZipFile(annotations_zip_2017, "r") as zip_dir:
                zip_dir.extractall(self.root)
        images_local = self._get_local_images_zip()
        images_gcs = f"{cloud_path}/{self.split}2017.zip"
        if not os.path.exists(images_local):
            logger.info(
                f"no zip file for images for {self.split} found,"
                f" will download"
            )
            client.download(
                local_path=self.root, bucket=const.GCS_BUCKET, key=images_gcs,
            )
            with zipfile.ZipFile(images_local, "r") as zip_dir:
                zip_dir.extractall(self.root)



import os
from datetime import datetime

TIMESTAMP_SUFFIX = datetime.now().strftime("%Y%m%d-%H%M%S")
PROJECT_ROOT = os.path.dirname(os.path.dirname(__file__))

# Default for tensorboard logs, checkpoints and metrics
DEFAULT_KFP_METRICS_DIR = os.path.join(
    PROJECT_ROOT, "metrics", TIMESTAMP_SUFFIX
)
DEFAULT_KFP_METRICS_FILENAME = "mlpipeline-metrics.json"

DEFAULT_TENSORBOARD_LOG_DIR = os.path.join(
    PROJECT_ROOT, "runs", TIMESTAMP_SUFFIX
)
DEFAULT_CHECKPOINT_DIR = os.path.join(
    PROJECT_ROOT, "checkpoints", TIMESTAMP_SUFFIX
)

GCS_BUCKET = "thea-dev"
GCS_BASE_STR = "gs://"
HTTP_URL_BASE_STR = "http://"
HTTPS_URL_BASE_STR = "https://"

# This is a hack on yacs config system, as it does not allow null values
# in configs. They are working on supporting null values in config
# https://github.com/rbgirshick/yacs/pull/18.
NULL_STRING = "None"

# Root directory of all datasets
# We assume the datasets are stored in the following structure:
# data_root/
#   cityscapes/
#   kitti/
#   nuscenes/
#   synthetic/
#   ...
DEFAULT_DATA_ROOT = "/data"
SYNTHETIC_SUBFOLDER = "synthetic"
DEFAULT_PUBLIC_DATASET = "synthetic"

# Default Unity Project ID where USim jobs was executed
DEFAULT_PROJECT_ID = "474ba200-4dcc-4976-818e-0efd28efed30"
USIM_API_ENDPOINT = "https://api.simulation.unity3d.com"

# Default Timing text for codetiming.Timer decorator
TIMING_TEXT = "[{name}] elapsed time: {:0.4f} seconds."

# Click CLI context settings
CONTEXT_SETTINGS = {
    "help_option_names": ["-h", "--help"],
    "show_default": True,
    "ignore_unknown_options": True,
    "allow_extra_args": True,
}
DEFAULT_DATASET_VERSION = "latest"



import torch


def create_loader(
    dataset, *, dryrun=False, batch_size=1, num_workers=0, collate_fn=None
):
    """Create data loader from dataset

    Note: The data loader here is a pytorch data loader object which does not
    assume tensor_type to be pytorch tensor. We only require input dataset to
    support `__getitem__` and `__len__` mothod to iterate over items in
    the dataset.

    Since `collate_fn` method in `torch.utils.data.DataLoader` behave
    differently when automatic batching is on, we might need to override
    this method. If `create_loader` method became too complicated in order to
    support different estimators, we might expect different estimator to
    have their own create_loader method.

    https://pytorch.org/docs/stable/data.html#working-with-collate-fn

    Args:
        dataset (Dataset): dataset object derived from
            `datasetinsights.data.datasets.Dataset` class.
        dryrun (bool): indicator whether to use a very small subset of the
            dataset. This subset is useful to make sure we can quickly run
            estimator without loading the whole dataset. (default: False)
        batch_size (int): how many samples per batch to load (default: 1)
        num_workers (int): number of parallel workers used for data loader.
            Set to `0` to run on a single thread (instead of `1` which might
            introduce overhead). (default: 0)
    Returns:
        `torch.utils.data.DataLoader` object as data loader
    """
    if dataset.split == "train":
        sampler = torch.utils.data.RandomSampler(dataset)
        drop_last = True
    else:
        sampler = torch.utils.data.SequentialSampler(dataset)
        drop_last = False

    if dryrun:
        sampler = torch.utils.data.SubsetRandomSampler(range(batch_size))

    data_loader = torch.utils.data.DataLoader(
        dataset,
        batch_size=batch_size,
        sampler=sampler,
        num_workers=num_workers,
        drop_last=drop_last,
        collate_fn=collate_fn,
    )
    return data_loader



from .average_log10_error import AverageLog10Error
from .average_precision_2d import (
    AveragePrecision,
    AveragePrecisionIOU50,
    MeanAveragePrecisionAverageOverIOU,
    MeanAveragePrecisionIOU50,
)
from .average_recall_2d import AverageRecall, MeanAverageRecallAverageOverIOU
from .average_relative_error import AverageRelativeError
from .base import EvaluationMetric
from .iou import IoU
from .root_mean_square_error import RootMeanSquareError
from .threshold_accuracy import ThresholdAccuracy

__all__ = [
    "EvaluationMetric",
    "IoU",
    "AverageRelativeError",
    "AverageLog10Error",
    "AveragePrecision",
    "AveragePrecisionIOU50",
    "AverageRecall",
    "MeanAveragePrecisionAverageOverIOU",
    "MeanAveragePrecisionIOU50",
    "MeanAverageRecallAverageOverIOU",
    "RootMeanSquareError",
    "ThresholdAccuracy",
]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['datasetinsights',
 'datasetinsights.commands',
 'datasetinsights.datasets',
 'datasetinsights.datasets.dummy',
 'datasetinsights.datasets.protos',
 'datasetinsights.datasets.unity_perception',
 'datasetinsights.estimators',
 'datasetinsights.evaluation_metrics',
 'datasetinsights.io',
 'datasetinsights.io.downloader',
 'datasetinsights.stats',
 'datasetinsights.stats.visualization']

package_data = \
{'': ['*'],
 'datasetinsights': ['configs/*'],
 'datasetinsights.stats.visualization': ['font/*']}

install_requires = \
['click>=7.1.2,<8.0.0',
 'codetiming>=1.2.0,<2.0.0',
 'cython>=0.29.14,<0.30.0',
 'dash==1.12.0',
 'dask[complete]>=2.14.0,<3.0.0',
 'google-cloud-storage>=1.24.1,<2.0.0',
 'jupyter>=1.0.0,<2.0.0',
 'kornia>=0.1.4,<0.2.0',
 'numpy>=1.17,<1.18',
 'nuscenes-devkit>=1.0.2,<1.0.3',
 'pandas>=1.0.1,<2.0.0',
 'plotly>=4.4.1,<5.0.0',
 'pycocotools>=2.0.0,<3.0.0',
 'pyquaternion>=0.9.5,<0.10.0',
 'pytorch-ignite>=0.3.0,<0.4.0',
 'tensorboardx>=2.0,<3.0',
 'tensorflow>=2.2.0,<3.0.0',
 'torch>=1.4.0,<1.5.0',
 'torchvision>=0.5,<0.6',
 'tqdm>=4.45.0,<5.0.0',
 'yacs>=0.1.6,<0.2.0']

entry_points = \
{'console_scripts': ['datasetinsights = datasetinsights.main:entrypoint']}

setup_kwargs = {
    'name': 'datasetinsights',
    'version': '0.2.0b6',
    'description': 'Synthetic dataset insights.',
    'long_description': "Dataset Insights\n================\nThis repo enables users to understand their synthetic datasets by exposing the metrics collected when the dataset\nwas created e.g. object count, label distribution, etc. The easiest way to use Dataset Insights is\nto run our jupyter notebook provided in our docker image `unitytechnologies/datasetinsights`\n\nRequirements\n============\n\nThe Dataset Insight notebooks assume that the user has already generated a synthetic dataset using the Unity Perception package.\nTo learn how to create a synthetic dataset using Unity please see the\n[perception documentation](https://github.com/Unity-Technologies/com.unity.perception).\n\n\n## Running the Dataset Insights Jupyter Notebook Locally\nYou can either run the notebook by installing our python package or by using our docker image.\n\n### Running a Notebook Locally Using Docker\n\n#### Requirements\n[Docker](https://docs.docker.com/get-docker/) installed.\n\n#### Steps\n1. Run notebook server using docker\n\n```bash\ndocker run \\\n  -p 8888:8888 \\\n  -v $HOME/data:/data \\\n  -t unitytechnologies/datasetinsights:latest\n```\nThis command mounts directory `$HOME/data` in your local filesystem to `/data` inside the container.\nIf you are loading a dataset generated locally from a Unity app, replace this path with the root of your app's persistent data folder.\n\nExample persistent data paths from [SynthDet](https://github.com/Unity-Technologies/synthdet):\n* OSX: `~/Library/Application\\ Support/UnityTechnologies/SynthDet`\n* Linux: `$XDG_CONFIG_HOME/unity3d/UnityTechnologies/SynthDet`\n* Windows: `%userprofile%\\AppData\\LocalLow\\UnityTechnologies\\SynthDet`\n\n\n2. Go to `http://localhost:8888` in a web browser to open the Jupyter browser.\n3. Open and run the example notebook in `/datasetinsights/notebooks/` or create your own.\n   (todo replace docker container gcr.io/unity-ai-thea-test/thea with public links)\n\n## Running a Dataset Insights Jupyter Notebook via Google Cloud Platform (GCP)\n- To run the notebook on GCP's AI platform follow\n[these instructions](https://cloud.google.com/ai-platform/notebooks/docs/custom-container) and use the container `unitytechnologies/datasetinsights:latest`\n- Alternately, to run the notebook on kubeflow follow [these steps](https://www.kubeflow.org/docs/notebooks/setup/)\n\n### Download Dataset from Unity Simulation\n\n[Unity Simulation](https://unity.com/products/simulation) provides a powerful platform for running simulations at large scale. You can use the provided cli script to download Perception datasets generated in Unity Simulation:\n\n```bash\npython -m datasetinsights.scripts.usim_download \\\n  --data-root=$HOME/data \\\n  --run-execution-id=<run-execution-id> \\\n  --auth-token=<xxx>\n```\n\nThe `auth-token` can be generated using the Unity Simulation [CLI](https://github.com/Unity-Technologies/Unity-Simulation-Docs/blob/master/doc/cli.md#usim-inspect-auth). This script will download the synthetic dataset for the requested [run-execution-id](https://github.com/Unity-Technologies/Unity-Simulation-Docs/blob/master/doc/cli.md#argument-descriptions).\n\nIf the `--include-binary` flag is present, the images will also be downloaded. This might take a long time, depending on the size of the generated dataset.\n\n### Download SynthDet Dataset\n\nDownload SynthDet public dataset from GCS, including GroceriesReal and Synthetic dataset. You can use the provided cli script to download public dataset to reproduce our work.\n\nHere is the command line for GroceriesReal dataset download:\n\n```bash\npython -m datasetinsights.scripts.public_download \\\n  --name=GroceriesReal \\\n  --data-root=$HOME/data \\\n```\n",
    'author': 'Unity AI Perception Team',
    'author_email': 'perception@unity3d.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import os

import torch.distributed as dist


def is_master():
    rank = int(os.getenv("RANK", 0))
    return rank == 0


def is_dist_avail_and_initialized():
    return dist.is_available() and dist.is_initialized()


def get_world_size():
    """
    Returns: number of available devices
    """
    if not is_dist_avail_and_initialized():
        return 1
    return dist.get_world_size()



""" Save estimator checkpoints
"""
import logging
import os
import tempfile

import datasetinsights.constants as const
from datasetinsights.io.download import download_file
from datasetinsights.io.gcs import GCSClient
from datasetinsights.torch_distributed import is_master

logger = logging.getLogger(__name__)

DEFAULT_SUFFIX = "estimator"


class EstimatorCheckpoint:
    """Saves and loads estimator checkpoints.

    Assigns estimator checkpoint writer according to log_dir which is
    responsible for saving estimators. Writer can be a GCS or local writer.
    Assigns loader which is responsible for loading estimator from a given
    path. Loader can a local, GCS or HTTP loader.

    Args:
        estimator_name (str): name of the estimator
        checkpoint_dir (str): Directory where checkpoints are stored
        distributed (bool): boolean to determine distributed training

    Attributes:
        checkpoint_dir (str): Directory where checkpoints are stored
        distributed (bool): boolean to determine distributed training

    """

    def __init__(self, estimator_name, checkpoint_dir, distributed):
        self.distributed = distributed
        self._writer = self._create_writer(checkpoint_dir, estimator_name)

    @staticmethod
    def _create_writer(checkpoint_dir, estimator_name):
        """Creates writer object for saving checkpoints based on log_dir.

        Args:
            checkpoint_dir: Directory where checkpoints are stored
            estimator_name: Name of the estimator.

        Returns:
            Writer object (GCS or Local).
        """

        if checkpoint_dir is None:
            checkpoint_dir = const.DEFAULT_CHECKPOINT_DIR
            if not os.path.exists(checkpoint_dir):
                os.makedirs(checkpoint_dir)

        if checkpoint_dir.startswith(const.GCS_BASE_STR):
            writer = GCSEstimatorWriter(checkpoint_dir, estimator_name)
        elif os.path.isdir(checkpoint_dir):
            writer = LocalEstimatorWriter(checkpoint_dir, estimator_name)
        else:
            raise ValueError(
                f"Can not use {checkpoint_dir} as checkpoint directory."
            )

        return writer

    @staticmethod
    def _get_loader_from_path(path):
        """Gives loader method to load an estimator from a given path.

        Arguments:
            path: Path of the estimator. Can be local, GCS or HTTP path.

        Returns:
            Loader method for loading estimator according to the given path.
        """
        if path.startswith((const.HTTP_URL_BASE_STR, const.HTTPS_URL_BASE_STR)):
            method = load_from_http
        elif path.startswith(const.GCS_BASE_STR):
            method = load_from_gcs
        elif os.path.isfile(path):
            method = load_local
        else:
            raise ValueError(
                f"Given path: {path}, is either invalid or not supported."
                f"Currently supported path are local path (/path/to/file), "
                f"GCS path (gs://) and HTTP url (http:// or https://) path"
            )

        return method

    def save(self, estimator, epoch):
        """Save estimator to the log_dir.

        Args:
            estimator (datasetinsights.estimators.Estimator):
            datasetinsights estimator object.
            epoch (int): Epoch number.

        """
        if self.distributed and not is_master():
            return
        self._writer.save(estimator=estimator, epoch=epoch)

    def load(self, estimator, path):
        """Loads estimator from given path.

        Path can be either a local path or GCS path or HTTP url.

        Args:
            estimator (datasetinsights.estimators.Estimator):
            datasetinsights estimator object
            path (str): path of estimator

        """
        if self.distributed and not is_master():
            return

        load_method = self._get_loader_from_path(path)
        load_method(estimator, path)


class LocalEstimatorWriter:
    """Writes (saves) estimator checkpoints locally.

    Args:
        dirname (str): Directory where estimator is to be saved.
        prefix (str): Filename prefix of the checkpoint files.
        suffix (str): Filename suffix of the checkpoint files.
        create_dir (bool): Flag for creating new directory. Default: True.

    Attributes:
        dirname (str): directory name of where checkpoint files are stored
        prefix (str): filename prefix of the checkpoint files
        suffix (str): filename suffix of the checkpoint files

    """

    def __init__(
        self, dirname, prefix, *, suffix=DEFAULT_SUFFIX, create_dir=True
    ):
        self.dirname = dirname
        self.prefix = prefix
        self.suffix = suffix
        self.is_master = is_master()
        if create_dir:
            if not os.path.exists(dirname):
                os.makedirs(dirname)

        if not os.path.exists(dirname):
            raise ValueError(f"Directory path '{dirname}' is not found.")

    def save(self, estimator, epoch=None):
        """Save estimator to locally to log_dir.

        Args:
            estimator (datasetinsights.estimators.Estimator):
                datasetinsights estimator object.
            epoch (int): The current epoch number. Default: None

        Returns:
            Full path to the saved checkpoint file.

        """
        if epoch:
            filename = ".".join([self.prefix, f"ep{epoch}", self.suffix])
        else:
            filename = ".".join([self.prefix, self.suffix])
        path = os.path.join(self.dirname, filename)

        logger.debug(f"Saving estimator to {path}")
        estimator.save(path)

        return path


class GCSEstimatorWriter:
    """Writes (saves) estimator checkpoints on GCS.

    Args:
        cloud_path (str): GCS cloud path (e.g. gs://bucket/path/to/directoy)
        prefix (str): filename prefix of the checkpoint files
        suffix (str): filename suffix of the checkpoint files

    """

    def __init__(self, cloud_path, prefix, *, suffix=DEFAULT_SUFFIX):
        self._tempdir = tempfile.TemporaryDirectory().name
        self._client = GCSClient()
        self.cloud_path = cloud_path
        self._writer = LocalEstimatorWriter(
            self._tempdir, prefix, create_dir=True, suffix=suffix
        )

    def save(self, estimator, epoch=None):
        """Save estimator to checkpoint files on GCS.

        Args:
            estimator (datasetinsights.estimators.Estimator):
                datasetinsights estimator object.
            epoch (int): the current epoch number. Default: None

        Returns:
            Full GCS cloud path to the saved checkpoint file.

        """
        path = self._writer.save(estimator, epoch)
        filename = os.path.basename(path)
        full_cloud_path = os.path.join(self.cloud_path, filename)
        logger.debug(f"Copying estimator from {path} to {full_cloud_path}")
        self._client.upload(local_path=path, url=full_cloud_path)
        return full_cloud_path


def load_local(estimator, path):
    """Loads estimator checkpoints from a local path."""
    estimator.load(path)

    return path


def load_from_gcs(estimator, full_cloud_path):
    """Load estimator from checkpoint files on GCS.

    Args:
        estimator (datasetinsights.estimators.Estimator):
            datasetinsights estimator object.
        full_cloud_path: full path to the checkpoint file

    """
    filename = os.path.basename(full_cloud_path)
    with tempfile.TemporaryDirectory() as temp_dir:
        path = os.path.join(temp_dir, filename)
        logger.debug(f"Downloading estimator from {full_cloud_path} to {path}")
        client = GCSClient()
        client.download(local_path=temp_dir, url=full_cloud_path)
        estimator.load(path)


def load_from_http(estimator, url):
    """Load estimator from checkpoint files on GCS.

    Args:
        estimator (datasetinsights.estimators.Estimator):
            datasetinsights estimator object.
        url: URL of the checkpoint file

    """
    with tempfile.TemporaryDirectory() as temp_dir:
        logger.debug(f"Downloading estimator from {url} to {temp_dir}")
        path = download_file(source_uri=url, dest_path=temp_dir)
        logger.debug(f"Loading estimator from {path}")
        estimator.load(path)



""" IoU evaluation metrics
"""
from ignite.metrics import ConfusionMatrix

from .base import EvaluationMetric


class IoU(EvaluationMetric):
    """ Intersection over Union (IoU) metric per class

    The metric is defined for a pair of grayscale semantic segmentation images.

    Args:
        num_classes: number of calsses in the ground truth image
        output_transform: function that transform output pair of images

    Attributes:
        cm (ignite.metrics.ConfusionMatrix): pytorch ignite confusion matrix
        object.
    """

    def __init__(self, num_classes, output_transform=lambda x: x):
        self.cm = ConfusionMatrix(
            num_classes=num_classes,
            average=None,
            output_transform=output_transform,
        )

    def reset(self):
        self.cm.reset()

    def update(self, output):
        self.cm.update(output)

    def compute(self):
        cm = self.cm.compute()
        iou = cm.diag() / (cm.sum(dim=1) + cm.sum(dim=0) - cm.diag() + 1e-15)

        return iou



import logging

import click
from yacs.config import CfgNode as CN

import datasetinsights.constants as const
from datasetinsights.estimators.base import create_estimator

logger = logging.getLogger(__name__)


@click.command(
    help="Start model training (and optionally validation) tasks.",
    context_settings=const.CONTEXT_SETTINGS,
)
@click.option(
    "-c",
    "--config",
    type=click.STRING,
    required=True,
    help="Path to the config estimator yaml file.",
)
@click.option(
    "-t",
    "--train-data",
    type=click.Path(exists=True, file_okay=False),
    required=True,
    help="Directory on localhost where train dataset is located.",
)
@click.option(
    "-e",
    "--val-data",
    type=click.Path(exists=True, file_okay=False),
    default=None,
    help="Directory on localhost where validation dataset is located.",
)
@click.option(
    "-p",
    "--checkpoint-file",
    type=click.STRING,
    default=None,
    help=(
        "URI to a checkpoint file. If specified, model will load from "
        "this checkpoint and resume training."
    ),
)
@click.option(
    "-l",
    "--tb-log-dir",
    type=click.STRING,
    default=const.DEFAULT_TENSORBOARD_LOG_DIR,
    help=(
        "Path to the directory where tensorboard events should be stored. "
        "This Path can be GCS URI (e.g. gs://<bucket>/runs) or full path "
        "to a local directory."
    ),
)
@click.option(
    "-p",
    "--checkpoint-dir",
    type=click.STRING,
    default=const.DEFAULT_CHECKPOINT_DIR,
    help=(
        "Path to the directory where model checkpoint files should be stored. "
        "This Path can be GCS URI (e.g. gs://<bucket>/checkpoints) or "
        "full path to a local directory."
    ),
)
@click.option(
    "-w",
    "--workers",
    type=click.INT,
    default=0,
    help=(
        "Number of multiprocessing workers for loading datasets. "
        "Set this argument to 0 will disable multiprocessing which is "
        "recommended when running inside a docker container."
    ),
)
@click.option(
    "--no-cuda",
    is_flag=True,
    default=False,
    help=(
        "Force to disable CUDA. If CUDA is available and this flag is False, "
        "model will be trained using CUDA."
    ),
)
@click.option(
    "--no-val",
    is_flag=True,
    default=False,
    help="Force to disable validations.",
)
def cli(
    config,
    train_data,
    val_data,
    checkpoint_file,
    tb_log_dir,
    checkpoint_dir,
    workers,
    no_cuda,
    no_val,
):
    ctx = click.get_current_context()
    logger.debug(f"Called train command with parameters: {ctx.params}")
    logger.debug(f"Override estimator config with args: {ctx.args}")

    config = CN.load_cfg(open(config, "r"))

    estimator = create_estimator(
        name=config.estimator,
        config=config,
        checkpoint_file=checkpoint_file,
        tb_log_dir=tb_log_dir,
        checkpoint_dir=checkpoint_dir,
        no_cuda=no_cuda,
        no_val=no_val,
        workers=workers,
    )

    estimator.train(train_data=train_data, val_data=val_data)



from abc import ABCMeta, abstractmethod


class EvaluationMetric(metaclass=ABCMeta):
    """Abstract base class for metrics.
    """

    COMPUTE_TYPE = ""

    @staticmethod
    def create(name, **kwargs):
        """Create a new instance of the metric subclass

        Args:
            name (str): unique identifier for a metric subclass
            config (dict): parameters specific to each metric subclass
                used to create a metric instance

        Returns:
            an instance of the specified metric subclass
        """
        metric_cls = EvaluationMetric.find(name)

        return metric_cls(**kwargs)

    @staticmethod
    def find(name):
        """Find EvaluationMetric subclass based on the given name

        Args:
            name (str): unique identifier for a metric subclass

        Returns:
            a label of the specified metric subclass
        """
        metric_classes = EvaluationMetric.__subclasses__()
        metric_names = [d.__name__ for d in metric_classes]
        if name in metric_names:
            metric_cls = metric_classes[metric_names.index(name)]
            return metric_cls
        else:
            raise NotImplementedError(f"Unknown Metric class {name}!")

    @abstractmethod
    def reset(self):
        raise NotImplementedError("Subclass needs to implement this method")

    @abstractmethod
    def update(self, output):
        raise NotImplementedError("Subclass needs to implement this method")

    @abstractmethod
    def compute(self):
        raise NotImplementedError("Subclass needs to implement this method")
